#include "core.h"
#include "customassert.h"
#include "allocator.h"
#include "array.h"
#include "simplestring.h"
#include "packstream.h"
#include "modmanager.h"
#include "hashedstring.h"
#include "file.h"
#include "configmanager.h"
#include "filestream.h"
#include "frameworkutil.h"
#include "printmanager.h"

ModManager* ModManager::m_Instance = NULL;

ModManager::ModManager()
:	m_Mods()
{
}

ModManager::~ModManager()
{
	m_Mods.Clear();
}

ModManager* ModManager::GetInstance()
{
	if ( !m_Instance )
	{
		m_Instance = new ModManager();
	}
	return m_Instance;
}

void ModManager::DeleteInstance()
{
	if ( m_Instance )
	{
		delete( m_Instance );
	}
}

bool ModManager::InitializeMods( const char* ModCfgName )
{
	if( FileUtil::Exists( ModCfgName ) )
	{
		const FileStream ModCfgStream = FileStream( ModCfgName, FileStream::EFM_Read );
		ConfigManager::Load( ModCfgStream );

		SimpleString Prefix = FileUtil::StripFilename( ModCfgName ) + "/";

		STATICHASH( mEldritchMods );
		STATICHASH( NumMods );
		uint NumMods = ConfigManager::GetInt( sNumMods, 0, smEldritchMods );

		for( uint ModIndex = 0; ModIndex < NumMods; ++ModIndex )
		{
			const char* ModName = ConfigManager::GetSequenceString( "Mods%d", ModIndex, NULL, smEldritchMods );
			if ( ModName )
			{
				AddMod( (Prefix + ModName + "/").CStr() );
			}
		}
		return true;
	}
	return false;
}

bool ModManager::AddMod( const char* ModPath )
{
	if ( !FileUtil::IsDir( ModPath ) )
	{
		PRINTF( "Could not add Mod from path %s: Path does not exist\n", ModPath );
		return false;
	}
	SimpleString ModConfig = SimpleString( ModPath ) + "mod.config";
	if ( ! FileUtil::Exists( ModConfig.CStr() ) )
	{
		PRINTF( "Could not add Mod from path %s: mod.config does not exist\n", ModPath );
		return false;
	}
	ConfigManager::PushEnvironment();
	ConfigManager::Load( FileStream( ModConfig.CStr(), FileStream::EFM_Read ) );
	PackStream::StaticAddPackageSpace( ModPath );

	STATICHASH( ModName );
	const char* ModName = ConfigManager::GetString( sModName );

	ASSERT( ModName );

	SimpleString ModContext = SimpleString( "Mod." ) + ModName;
	MAKEHASH( ModContext );

	// TODO: expand with other mod properties (see mod.config)

	STATICHASH( NumPackages );
	uint NumPackages = ConfigManager::GetInt( sNumPackages, 0, sModContext );
	for ( uint i = 0; i < NumPackages; ++i )
	{
		SimpleString PkgPath = SimpleString( ModPath ) + ConfigManager::GetSequenceString( "Package%d", i, NULL, sModContext );
		PackStream::StaticAddPackageFile( PkgPath.CStr() );
	}

	STATICHASH( NumConfigFiles );
	uint NumConfigFiles = ConfigManager::GetInt( sNumConfigFiles, 0, sModContext );
	for ( uint i = 0; i < NumConfigFiles; ++i )
	{
		LoadConfigFile( ConfigManager::GetSequenceString( "ConfigFile%d", i, NULL, sModContext ) );
	}

	// clean up
	PackStream::StaticMakeCurrentPackageSpace();
	ConfigManager::RemoveContext( "" );
	ConfigManager::MergeEnvironment();

	ModDescription& TheMod = m_Mods.PushBack( ModDescription( ModName, ModPath ) );

	TheMod.m_Version = ConfigManager::GetString( "Version", "", sModContext );
	TheMod.m_WebsiteURL = ConfigManager::GetString( "WebsiteURL", "", sModContext );
	TheMod.m_DownloadURL = ConfigManager::GetString( "DownloadURL", "", sModContext );

	PRINTF( "Added Mod \"%s\" from path %s\n", ModName, ModPath );

	return true;
}

bool ModManager::LoadConfigFile( const char* ConfigFile )
{
	const char* Prefix = PackStream::StaticGetCurrentPrefix();

	bool Loaded = false;
	// find name of uncompiled config file
	SimpleString FileName = ConfigFile;
	int Where = FileName.Find( ".ccf" );
	if ( Where >= 0 and FileName.Length() - Where == 4 )
	{
		FileName = FileName.Replace( Where, ".config" );
	}
	else
	{ 	// special case for syncer config file...
		Where = FileName.Find( ".pcf" );
		if ( Where >= 0 and FileName.Length() - Where == 4 )
		{
			FileName = FileName.Replace( Where, ".syncerconfig" );
			if ( FileUtil::Exists( FileName.CStr() ) )
			{
				ConfigManager::Load( FileStream( FileName.CStr(), FileStream::EFM_Read ) );
			}
		}
		else
		{
			// either ConfigFile already is the uncompiled version, or something we don't understand.
			FileName = "";
		}
	}

	// load compiled config file from packstream
	if ( PackStream::StaticFileExists( ConfigFile, true ) )
	{
		ConfigManager::Load( PackStream( ConfigFile, PackStream::PSO_PackedOnly ) );
		PackStream::StaticMakeCurrentPackageSpace( Prefix );
		Loaded = true;
	}
	// load uncompiled config file from packstream
	if ( ( FileName != "" ) && PackStream::StaticFileExists( FileName.CStr(), true ) )
	{
		ConfigManager::Load( PackStream( FileName.CStr(), PackStream::PSO_PackedOnly ) );
		PackStream::StaticMakeCurrentPackageSpace( Prefix );
		Loaded = true;
	}

	SimpleString RealConfigFile = SimpleString( Prefix ) + ConfigFile;
	SimpleString RealFileName = SimpleString( Prefix ) + FileName;

	// try compiled config file from disc
	if ( FileUtil::Exists( RealConfigFile.CStr() ) )
	{
		ConfigManager::Load( FileStream( RealConfigFile.CStr(), FileStream::EFM_Read ) );
		PackStream::StaticMakeCurrentPackageSpace( Prefix );
		Loaded = true;
	}
	// also try uncompiled config file
	if ( ( FileName != "" ) && FileUtil::Exists( RealFileName.CStr() ) )
	{
		ConfigManager::Load( FileStream( RealFileName.CStr(), FileStream::EFM_Read ) );
		PackStream::StaticMakeCurrentPackageSpace( Prefix );
		Loaded = true;
	}
	return Loaded;
}

PackStream ModManager::OpenPackStream( const char* FileName )
{
	for ( int i = m_Mods.Size() - 1; i >= 0; --i )
	{
		ModDescription& TheMod = m_Mods[ i ];
		if ( PackStream::StaticMakeCurrentPackageSpace( TheMod.m_Space ) )
		{
			if ( PackStream::StaticFileExists( FileName ) )
			{
				return PackStream( FileName, PackStream::PSO_PackedLast );
				// TODO PackStream::StaticMakeCurrentPackageSpace();
			}
		}
	}
	// if the file was not found in a mod, open it from eldritch
	PackStream::StaticMakeCurrentPackageSpace();
	return PackStream( FileName, PackStream::PSO_PackedLast );
}
